module SymmetryAxis
  # The algorithm, based on principle, every point have a pair.
  # It uses the first got point and iterates over +(N - 1)+
  # other points to find a line,
  # that another points lie by twos on lines parallel to.
  # The perpendicular line to this line is the symmetry axis.
  class Finder
    attr_reader :line

    def initialize(point_array)
      @input = point_array
      return if @input.size.odd?
      @impossible_angles = []
      @two_points_cursor = 1
      loop do
        return unless take_two_points # step 1
        break if points_are_on_parallel_lines # step 2
      end
      find_axis # step 3
    end

    private

    def take_two_points
      @two_points = [@input.first]
      (@two_points_cursor...(@input.size)).each do |i|
        rad = Line.for_points(@two_points[0], @input[i]).angle_rad
        next if impossible?(rad)
        choose_point i
        return true
      end
      false
    end

    def choose_point(i)
      @two_points_cursor = i + 1
      @two_points[1] = @input[i]
      @two_points_line = Line.for_points(@two_points[0], @input[i])
    end

    def impossible?(rad)
      return false if @impossible_angles.empty?
      impossible_rad = @impossible_angles.min_by { |e| (e - rad).abs }
      (rad - impossible_rad).abs < (2 * Line.EPS)
    end

    def points_are_on_parallel_lines
      other_points = @input.select { |p| !@two_points.include?(p) }
      @axis_perpendicular_points = @two_points.dup
      check_perpendiculars(other_points)
    end

    def check_perpendiculars(points)
      loop do
        break if points.empty?
        line = @two_points_line.parallel(*points.first)
        size0 = points.size
        on_line = points.select { |p2| line.on_me?(*p2) }
        points -= on_line
        return false if size0 - points.size < 2
        save_perpendicular_points(on_line, line)
      end
      true
    end

    def save_perpendicular_points(points, line)
      return if line != @two_points_line
      points += @two_points
      @axis_perpendicular_points = pppair(points, line)
    end

    def pppair(points, line)
      if line.vertical?
        [points.min_by { |p| p[1] }, points.max_by { |p| p[1] }]
      else
        [points.min_by { |p| p[0] }, points.max_by { |p| p[0] }]
      end
    end

    def find_axis
      p1, p2 = *@axis_perpendicular_points
      @line = Line.symmetryof(p1[0], p1[1], p2[0], p2[1])
    end
  end
end
