module SymmetryAxis
  # The lines here are defined by the coefficients of the equation:
  #     y = a + bx
  # But this equation can not describe vertical lines (infinite +b+).
  # So the vertical lines are defined by +x+.
  class Line
    attr_reader :a, :b, :x
    EPS, TOO_BIG_B = 0.000001, 100_000_000

    def initialize(x1, y1, x2, y2)
      @vertical = nil
      if x1 < x2
        resolve_coefs(x1, y1, x2, y2)
      else
        resolve_coefs(x2, y2, x1, y1)
      end
    end

    # Shorthand for constructor.
    # @return [Line]
    def self.for_points(p1, p2)
      Line.new(p1[0], p1[1], p2[0], p2[1])
    end

    def ==(other)
      self.class.aeq(self, other) && self.class.beq(self, other)
    end

    # @return [String]
    def to_s
      if !vertical?
        "(a=#{@a}, b=#{@b})"
      else
        "(x=#{@x})"
      end
    end

    # Is the line strictly vertical. This line could be created from
    # almost vertical pair of points (+b+ coefficient was just Too Big).
    #
    # If +true+, you cannot use +a+ and +b+ fields
    # (you can read +x+ field instead).
    def vertical?
      @vertical = (nil == @b) || (TOO_BIG_B <= @b.abs) if nil == @vertical
      @vertical
    end

    # Zero means three o'clock, +pi/2+ means noon.
    # @return [Float]
    def angle_rad
      return (Math::PI / 2) if vertical?
      Math.atan(@b)
    end

    # Zero means three o'clock, 90 means noon.
    # @return [Float]
    def angle_degrees
      angle_rad / Math::PI * 180
    end

    # Does the point lie on the line.
    def on_me?(x, y)
      if vertical?
        (@x - x).abs < EPS
      else
        f = @a + @b * x
        (f - y).abs < on_me_eps(x)
      end
    end

    def on_me_eps(x)
      e = EPS
      e += (EPS / @b).abs if @b != 0
      e += (EPS / x).abs if x != 0
      e
    end

    # Symmetry line of two points.
    # The perpendicular line drawn through the center
    # of the line segment between the points.
    # @return [Line]
    def self.symmetryof(x1, y1, x2, y2)
      line = Line.new(x1, y1, x2, y2)
      x, y = (x1 + x2) / 2, (y1 + y2) / 2
      if line.vertical?
        Line.new(0, y, 1, y)
      else
        r = line.angle_rad + (Math::PI / 2)
        return Line.new(x, 0, x, 1) if Math::PI / 2 == r.abs
        return Line.new(x, y, x + 1, y + Math.tan(r))
      end
    end

    # Draw a parallel line through the point.
    # @return [Line]
    def parallel(x, y)
      return Line.new(x, y, x, y + 1) if vertical?
      Line.new(x, y, x + 1, y + @b)
    end

    private :on_me_eps

    private

    def resolve_coefs(xmin, yxmin, xmax, yxmax)
      if 0 == (xmax - xmin).abs
        @b, @a = nil, yxmin
      else
        @b = (yxmax - yxmin) / Float(xmax - xmin)
        @a = yxmin - @b * xmin
      end
      @x, @a = xmin, 0 if vertical?
    end

    def self.aeq(o1, o2)
      (o1.a - o2.a).abs < (2 * EPS)
    end

    def self.beq(o1, o2)
      if o1.vertical? || o2.vertical?
        return false unless o1.vertical? && o2.vertical?
        (o1.x - o2.x).abs < (2 * EPS)
      else
        (o1.b - o2.b).abs < (2 * EPS)
      end
    end
  end
end
